import inspect
import re
import imp
import os.path
import sys
from types import BuiltinFunctionType, CodeType, GetSetDescriptorType, MappingProxyType, MethodDescriptorType, ModuleType, WrapperDescriptorType


def get_actual_func_globals(func) -> dict:
    code = func.__code__
    func_globals = func.__globals__.items()
    actual_globals = {}
    for glob in func_globals:
        if glob[0] in code.co_names:
            actual_globals.update({glob[0]: glob[1]})
    return actual_globals


def get_actual_code_fields(_code: CodeType) -> dict:
    code_dict = {}
    for member in inspect.getmembers(_code):
        if str(member[0]).startswith("co_"):
            code_dict.update({member[0]: member[1]})
    return code_dict


def get_actual_module_fields(module: ModuleType) -> dict:
    module_fields = {}
    module_members = inspect.getmembers(module)
    for mem in module_members:
        if not mem[0].startswith("__"):
            module_fields.update({mem[0]: mem[1]})
    return module_fields


def get_actual_class_fields(_class: type) -> dict:
    fields_dict = {}
    if _class == type:
        fields_dict.update({
            "__bases__": [],
        })
    else:
        mems = inspect.getmembers(_class)
        for mem in mems:
            if type(mem[1]) not in (
                WrapperDescriptorType,
                MethodDescriptorType,
                BuiltinFunctionType,
                MappingProxyType,
                GetSetDescriptorType
            ):
                if mem[0] not in (
                    "__mro__", "__base__", "__basicsize__",
                    "__class__", "__dictoffset__", "__name__",
                    "__qualname__", "__text_signature__", "__itemsize__",
                    "__flags__", "__weakrefoffset__"
                ):
                    fields_dict.update({mem[0]: mem[1]})
    return fields_dict




def is_std_lib_module(module: ModuleType):
    python_libs_path = sys.path[2]
    module_path = imp.find_module(module.__name__)[1]
    if module.__name__ in sys.builtin_module_names:
        return True
    elif python_libs_path in module_path:
        return True
    elif 'site-packages' in module_path:
        return True
    return False





import imp
import re
from types import CodeType, FunctionType, ModuleType

from ...dto import DTO, DTO_TYPES
from . import yaml_tokens
TOKEN_TYPES = yaml_tokens

class YamlParser():

    __tokens: list = []
    __cur_gaps_count: int = 0

    def _eat(self, token_types: tuple, gaps: bool = False) -> tuple:
        if gaps:
            self._skip_gaps()
        if len(self.__tokens):
            if self.__tokens[0][0] in token_types:
                return self.__tokens.pop(0)
        return ("", "")

    def _head_token(self, skip_gaps: bool = False) -> tuple:
        head = self.__tokens[0]
        if skip_gaps:
            for token in self.__tokens:
                if token[0] != TOKEN_TYPES.GAP:
                    head = token
                    break
        return head

    def _lex(self, source: str) -> list:
        tokens = []
        while len(source) > 0:
            for token in TOKEN_TYPES.TOKEN_REGEXPS.items():
                if token[0] == TOKEN_TYPES.EOF:
                    continue
                regexp_res = re.match(token[1], source)
                if regexp_res and regexp_res.start() == 0 and len(regexp_res.group(0)):
                    source = str(
                        source[regexp_res.end()-regexp_res.start():])
                    if token[0] in (TOKEN_TYPES.STR, TOKEN_TYPES.FIELD_STR):
                        string = regexp_res.group(0)
                        tokens.append((token[0], string.replace('"', "")))
                    elif token[0] == TOKEN_TYPES.NUMBER:
                        num = regexp_res.group(0)
                        num = float(num) if "." in num else int(num)
                        tokens.append((token[0], num))
                    elif token[0] == TOKEN_TYPES.BOOL:
                        _bool = regexp_res.group(0)
                        res = True if _bool == "true" else False
                        tokens.append((token[0], res))
                    else:
                        tokens.append((token[0],))
        tokens.append((TOKEN_TYPES.EOF,))
        return tokens

    def _dump_tokens(self):
        for token in self.__tokens:
            print(token)

    def _skip_field_name(self) -> str:
        self._skip_gaps()
        field_key = self._eat(TOKEN_TYPES.FIELD_STR)
        self._eat(TOKEN_TYPES.COLON)
        self._skip_gaps()
        return field_key[1]

    def _peek_gaps_count(self) -> int:
        for i in range(len(self.__tokens)):
            if self.__tokens[i][0] != TOKEN_TYPES.GAP:
                return i

    def _skip_gaps(self) -> int:
        i = 0
        while self._head_token()[0] == TOKEN_TYPES.GAP:
            self._eat(TOKEN_TYPES.GAP)
            i += 1
        return i

    def _push_gap(self):
        self.__cur_gaps_count += 1

    def _pop_gap(self):
        self.__cur_gaps_count -= 1

    def _parse_func_code(self) -> CodeType:
        self._push_gap()
        # fields
        self._skip_field_name()
        code_dict = self._parse()

        self._pop_gap()

        func_code = CodeType(
            int(code_dict["co_argcount"]),
            int(code_dict["co_posonlyargcount"]),
            int(code_dict["co_kwonlyargcount"]),
            int(code_dict["co_nlocals"]),
            int(code_dict["co_stacksize"]),
            int(code_dict["co_flags"]),
            bytes.fromhex(code_dict["co_code"]),
            tuple(code_dict["co_consts"]),
            tuple(code_dict["co_names"]),
            tuple(code_dict["co_varnames"]),
            str(code_dict["co_filename"]),
            str(code_dict["co_name"]),
            int(code_dict["co_firstlineno"]),
            bytes.fromhex(code_dict["co_lnotab"]),
            tuple(code_dict["co_freevars"]),
            tuple(code_dict["co_cellvars"]),
        )
        return func_code

    def _parse_func(self) -> any:
        self._push_gap()
        # name
        self._skip_field_name()
        func_name = self._parse()
        # globals
        self._skip_field_name()
        func_globals = self._parse()
        # code
        self._skip_field_name()
        func_code = self._parse()
        
        # print(func_name, func_globals, func_code)
        # exit()
        self._pop_gap()
        # print()
        # exit()

        func = FunctionType(func_code, func_globals, func_name)
        func.__globals__["__builtins__"] = __import__("builtins")
        return func

    def _parse_module(self) -> ModuleType:
        self._push_gap()
        # name
        self._skip_field_name()
        module_name = self._parse()
        # fields
        self._skip_field_name()
        module_fields = self._parse()

        self._pop_gap()

        module = None
        # first case mean that module is in python std or built in
        if module_fields == None:
            module = __import__(module_name)
        else: 
            module = imp.new_module(module_name)
            for field in module_fields.items():
                setattr(module,field[0],field[1])
        return module

    def _parse_class(self) -> type:
        self._push_gap()
        # name
        self._skip_field_name()
        class_name = self._parse()
        # fields
        self._skip_field_name()
        class_members_dict = self._parse()

        # print(class_members_dict)
        # exit()
        self._pop_gap()

        class_bases = (object,)
        if "__bases__" in class_members_dict:
            class_bases = tuple(class_members_dict["__bases__"])
        return type(class_name, class_bases, class_members_dict)

    def _parse_obj(self) -> object:
        self._push_gap()
        # class
        self._skip_field_name()
        _class = self._parse()
        # fields
        self._skip_field_name()
        fields_dict = self._parse()
        self._pop_gap()

        class_init = _class.__init__
        if callable(class_init):
            if class_init.__class__.__name__ == "function":
                delattr(_class, "__init__")
        obj = _class()
        obj.__init__ = class_init
        obj.__dict__ = fields_dict
        return obj

    def _parse_dict(self):
        _dict = {}
        start_gaps_count = self.__cur_gaps_count
        self._push_gap()
        while self._peek_gaps_count() == start_gaps_count and self._head_token()[0] != TOKEN_TYPES.EOF:
            co_key = self._skip_field_name()
            co_value = self._parse()
            _dict.update({co_key: co_value})
        self._pop_gap()
        return _dict

    def _parse_list(self) -> list:
        _list = []
        start_gaps_count = self.__cur_gaps_count
        self._push_gap()
        while self._peek_gaps_count() == start_gaps_count and self._head_token()[0] != TOKEN_TYPES.EOF:
            self._skip_gaps()
            self._eat(TOKEN_TYPES.ARR_DASH)
            element = self._parse()
            _list.append(element)
        self._pop_gap()
        return _list

    def _parse_primitive(self) -> any:
        token_type = self._head_token()[0]
        res = None
        if token_type == TOKEN_TYPES.NUMBER:
            res = self._eat(TOKEN_TYPES.NUMBER)[1]
        elif token_type == TOKEN_TYPES.STR:
            res = self._eat(TOKEN_TYPES.STR)[1]
        elif token_type == TOKEN_TYPES.BOOL:
            res = self._eat(TOKEN_TYPES.BOOL)[1]
        elif token_type == TOKEN_TYPES.NULL:
            self._eat(TOKEN_TYPES.NULL)
            res = None
        self._eat(TOKEN_TYPES.NEW_LINE)
        return res

    def _parse_dto(self):

        dto_type_key = self._eat(TOKEN_TYPES.FIELD_STR, gaps=True)
        self._eat(TOKEN_TYPES.COLON)
        dto_type_value = self._eat(TOKEN_TYPES.STR)
        self._eat(TOKEN_TYPES.NEW_LINE)

        res_dto = None

        # self._push_gap()
        if dto_type_key[1] == DTO.dto_type:
            if dto_type_value[1] == DTO_TYPES.DICT:
                res_dto = self._parse_dict()
            elif dto_type_value[1] == DTO_TYPES.FUNC:
                res_dto = self._parse_func()
            elif dto_type_value[1] == DTO_TYPES.CODE:
                res_dto = self._parse_func_code()
            elif dto_type_value[1] == DTO_TYPES.MODULE:
                res_dto = self._parse_module()
            elif dto_type_value[1] == DTO_TYPES.CLASS:
                res_dto = self._parse_class()
            elif dto_type_value[1] == DTO_TYPES.OBJ:
                res_dto = self._parse_obj()
        # self._pop_gap()
        return res_dto

    def _parse(self) -> any:
        head_token_type = self._head_token()[0]
        # print(head_token_type)
        if head_token_type == TOKEN_TYPES.LBRACKET:
            self._eat(TOKEN_TYPES.LBRACKET)
            self._eat(TOKEN_TYPES.RBRACKET)
            self._eat(TOKEN_TYPES.NEW_LINE)
            return []
        elif head_token_type in (TOKEN_TYPES.NEW_LINE, TOKEN_TYPES.FIELD_STR):
            self._eat(TOKEN_TYPES.NEW_LINE)
            if self._head_token(skip_gaps=True)[0] == TOKEN_TYPES.ARR_DASH:
                return self._parse_list()
            self._skip_gaps()
            return self._parse_dto()
        return self._parse_primitive()

    def parse(self, s: str) -> any:
        self.__tokens = self._lex(s)
        return self._parse()



import argparse
import os
import sys
import configparser

from ..serializer import create_serializer

class CLI:
    __parser = None

    def parse_args(self):
        self.__parser = argparse.ArgumentParser(
            description='Dmitry Pekutko formats serializer'
        )
        self.__parser.add_argument(
            "--source", type=str, help="File path to parse"
        )
        self.__parser.add_argument(
            "--to", type=str, help="File path to write"
        )
        self.__parser.add_argument(
            "--conf", type=str, help="Ignore args and use config file with path"
        )
        return self.__parser.parse_args()

    def exec_convert(self, cli_args):
        source_path, to_path = None, None
        # config strategy
        if cli_args.conf:
            source_path, to_path = self.__read_config(cli_args.conf)
        # cli args strategy
        else:
            source_path, to_path = cli_args.source, cli_args.to
        # execute convertation
        if source_path and to_path:
            if os.path.exists(source_path) and os.path.isfile(source_path):
                abs_source_path = os.path.abspath(source_path)
                abs_to_path = os.path.abspath(to_path)
                self.__convert(abs_source_path, abs_to_path)
            else:
                self.__parser.error(f'No such file: "{source_path}"')
        else:
            self.__parser.error(
                f'Invalid --source and --to flags: source="{source_path}", to="{to_path}"')

    def __read_config(self, config_path: str) -> tuple:
        config_path = os.path.abspath(config_path)
        if os.path.exists(config_path) and os.path.isfile(config_path):
            config = configparser.RawConfigParser()
            config.read(config_path)
            source_path = config.get("DEFAULT", "source_path")
            to_path = config.get("DEFAULT", "to_path")
            return source_path.replace('"',''), to_path.replace('"','')
        else:
            self.__parser.error(f'Invalid config file path: "{config_path}"')

    def __convert(self, source_file_path: str, to_file_path: str):
        source_format = os.path.splitext(source_file_path)[1][1:]
        to_format = os.path.splitext(to_file_path)[1][1:]
        if source_format != to_format:
            source_ser = create_serializer(source_format)
            to_ser = create_serializer(to_format)
            if source_ser and to_ser:
                obj = source_ser.load(source_file_path)
                to_ser.dump(obj, to_file_path)
        else:
            print('Identic formats, exit.')



from .serializer.json.JsonSerializer import JsonSerializer
from .serializer.toml.TomlSerializer import TomlSerializer
from .serializer.yaml.YamlSerializer import YamlSerializer
from .serializer.create_serializer import create_serializer



import imp
import inspect
import re
from types import CodeType, FunctionType, ModuleType

from ...dto import DTO, DTO_TYPES
from . import json_tokens
TOKEN_TYPES = json_tokens

class JsonParser():

    __tokens: list = []

    def _eat(self, token_types: tuple) -> tuple:
        if len(self.__tokens):
            if self.__tokens[0][0] in token_types:
                return self.__tokens.pop(0)
        return ("", "")

    def _head_token(self) -> tuple:
        if len(self.__tokens):
            return self.__tokens[0]

    def _lex(self, source: str) -> list:
        tokens = []
        while len(source) > 0:
            for token in TOKEN_TYPES.TOKEN_REGEXPS.items():
                if token[0] == TOKEN_TYPES.EOF:
                    continue
                try:
                    regexp_res = re.match(token[1], source)
                    if regexp_res.start() == 0:
                        source = str(
                            source[regexp_res.end()-regexp_res.start():]).strip()
                        if token[0] == TOKEN_TYPES.STR:
                            string = regexp_res.group(0)
                            tokens.append((token[0], string.replace('"', "")))
                        elif token[0] == TOKEN_TYPES.NULL:
                            tokens.append((token[0], None))
                        elif token[0] == TOKEN_TYPES.NUMBER:
                            num = regexp_res.group(0)
                            num = float(num) if "." in num else int(num)
                            tokens.append((token[0], num))
                        elif token[0] == TOKEN_TYPES.BOOL:
                            _bool = regexp_res.group(0)
                            res = True if _bool == "true" else False
                            tokens.append((token[0], res))
                        else:
                            tokens.append((token[0],))
                except:
                    None
        tokens.append((TOKEN_TYPES.EOF,))
        return tokens

    def _skip_field_name(self, comma: bool = False) -> str:
        if comma:
            self._eat(TOKEN_TYPES.COMMA)
        field_key = self._eat(TOKEN_TYPES.STR)
        self._eat(TOKEN_TYPES.COLON)
        return field_key[1]

    def _parse_func_code(self) -> CodeType:
        self._skip_field_name(comma=True)
        code_dict = self._parse()
        func_code = CodeType(
            int(code_dict["co_argcount"]),
            int(code_dict["co_posonlyargcount"]),
            int(code_dict["co_kwonlyargcount"]),
            int(code_dict["co_nlocals"]),
            int(code_dict["co_stacksize"]),
            int(code_dict["co_flags"]),
            bytes.fromhex(code_dict["co_code"]),
            tuple(code_dict["co_consts"]),
            tuple(code_dict["co_names"]),
            tuple(code_dict["co_varnames"]),
            str(code_dict["co_filename"]),
            str(code_dict["co_name"]),
            int(code_dict["co_firstlineno"]),
            bytes.fromhex(code_dict["co_lnotab"]),
            tuple(code_dict["co_freevars"]),
            tuple(code_dict["co_cellvars"]),
        )
        return func_code

    def _parse_func(self) -> any:
        # exit()
        # name
        self._skip_field_name()
        func_name = self._parse()
        # globals
        self._skip_field_name(comma=True)
        func_globals = self._parse()
        # exit()
        # code
        self._skip_field_name(comma=True)
        func_code = self._parse()

        func = FunctionType(func_code, func_globals, func_name)
        func.__globals__["__builtins__"] = __import__("builtins")
        return func

    def _parse_module(self) -> ModuleType:
        # name
        self._skip_field_name()
        module_name = self._parse()
        # fields
        self._skip_field_name(comma=True)
        module_fields = self._parse()

        module = None
        # first case mean that module is in python std or built in
        if module_fields == None:
            module = __import__(module_name)
        else: 
            module = imp.new_module(module_name)
            for field in module_fields.items():
                setattr(module,field[0],field[1])
        return module

    def _parse_class(self) -> type:
        # name
        self._skip_field_name()
        class_name = self._parse()
        # fields
        self._skip_field_name(comma=True)
        class_members_dict = self._parse()

        class_bases = (object,)
        if "__bases__" in class_members_dict:
            class_bases = tuple(class_members_dict["__bases__"])
        return type(class_name, class_bases, class_members_dict)

    def _parse_obj(self) -> object:
        # class
        self._skip_field_name()
        _class = self._parse()
        # fields
        self._skip_field_name(comma=True)
        fields_dict = self._parse()

        class_init = _class.__init__
        if callable(class_init):
            if class_init.__class__.__name__ == "function":
                delattr(_class, "__init__")
        obj = _class()
        obj.__init__ = class_init
        obj.__dict__ = fields_dict
        return obj

    def _parse_dict(self):
        _dict = {}
        is_first = True
        while self._head_token()[0] != TOKEN_TYPES.RBRACE:
            co_key = self._skip_field_name(not is_first)
            co_value = self._parse()
            _dict.update({co_key: co_value})
            is_first = False
        return _dict

    def _parse_list(self) -> list:
        self._skip_field_name()
        self._eat(TOKEN_TYPES.LBRACKET)
        _list = []
        is_first = True
        while self._head_token()[0] != TOKEN_TYPES.RBRACKET:
            # print(is_first)
            if not is_first:
                self._eat(TOKEN_TYPES.COMMA)
            _list.append(self._parse())
            is_first = False
        self._eat(TOKEN_TYPES.RBRACKET)
        return _list

    def _parse_primitive(self) -> any:
        token_type = self._head_token()[0]
        res = None
        if token_type == TOKEN_TYPES.NUMBER:
            res = self._eat(TOKEN_TYPES.NUMBER)[1]
        elif token_type == TOKEN_TYPES.STR:
            res = self._eat(TOKEN_TYPES.STR)[1]
        elif token_type == TOKEN_TYPES.NULL:
            res = self._eat(TOKEN_TYPES.NULL)[1]
        elif token_type == TOKEN_TYPES.BOOL:
            res = self._eat(TOKEN_TYPES.BOOL)[1]
        elif token_type in TOKEN_TYPES.LBRACKET:
            res = self._parse_list()
        if self._head_token()[0] == TOKEN_TYPES.COMMA:
            self._eat(TOKEN_TYPES.COMMA)
        return res

    def _parse_dto(self):
        self._eat(TOKEN_TYPES.LBRACE)

        if self._head_token()[0] == TOKEN_TYPES.RBRACE:
            self._eat(TOKEN_TYPES.RBRACE)
            return None

        dto_type_key = self._eat(TOKEN_TYPES.STR)
        self._eat(TOKEN_TYPES.COLON)
        dto_type_value = self._eat(TOKEN_TYPES.STR)
        self._eat(TOKEN_TYPES.COMMA)

        res_dto = None

        if dto_type_key[1] == DTO.dto_type:
            if dto_type_value[1] == DTO_TYPES.DICT:
                res_dto = self._parse_dict()
            elif dto_type_value[1] == DTO_TYPES.FUNC:
                res_dto = self._parse_func()
            elif dto_type_value[1] == DTO_TYPES.CODE:
                res_dto = self._parse_func_code()
            elif dto_type_value[1] == DTO_TYPES.MODULE:
                res_dto = self._parse_module()
            elif dto_type_value[1] == DTO_TYPES.CLASS:
                res_dto = self._parse_class()
            elif dto_type_value[1] == DTO_TYPES.OBJ:
                res_dto = self._parse_obj()
                
        self._eat(TOKEN_TYPES.RBRACE)
        return res_dto

    def _parse(self) -> any:
        head_token_type = self._head_token()[0]
        if head_token_type == TOKEN_TYPES.LBRACE:
            return self._parse_dto()
        return self._parse_primitive()

    def parse(self, s: str) -> any:
        self.__tokens = self._lex(s)
        return self._parse()
